from typing import List


VALID_KEYS = set(['C', 'C#', 'D', 'D#', 'E', 'F',
                 'F#', 'G', 'G#', 'A', 'A#', 'B'])


def int_to_str(key: int) -> str:
    '''
    key: 0 = C, 1 = C#, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
    '''

    if key < 0 or key > 11:
        raise ValueError('Invalid key')

    return ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'][key]


def str_to_int(key: str) -> int:
    '''
    key: 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'
    '''
    key = key.upper()
    key = flats_to_sharps(key)

    return ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'].index(key)


def flats_to_sharps(key: str) -> str:
    '''
    input key: 'C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B, C#', 'D#', 'F#', 'G#', 'A#, E#, B#'
    resulting key: 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'
    this function will convert all possible key names to the standard form
    '''

    if not isinstance(key, str):
        raise TypeError('Invalid key type')

    key = key[0].upper() + key[1:].lower()

    VALID_INPUTS = set(['C', 'C#', 'Db', 'D', 'D#', 'Eb', 'E', 'Fb', 'E#',
                       'F', 'F#', 'Gb', 'G', 'G#', 'Ab', 'A', 'A#', 'Bb', 'B', 'B#', 'Cb'])

    if key not in VALID_INPUTS:
        raise ValueError('Invalid key')

    flat_to_sharp_dict = {'Db': 'C#', 'Eb': 'D#',
                          'Gb': 'F#', 'Ab': 'G#', 'Bb': 'A#', 'Cb': 'B', 'Fb': 'E'}

    if key in flat_to_sharp_dict:
        return flat_to_sharp_dict[key]
    else:
        return key


class Note:
    note_str: str = None
    note_int: int = None
    octave: int = None  # TODO: implement octave

    def __init__(self, note_str: str = None, note_int: int = None):
        '''
        note_str: 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'
        #, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        note_int: 0 = C, 1 = C
        '''

        if note_str is not None:
            self.note_str = note_str
            self.note_int = str_to_int(note_str)
        elif note_int is not None:
            self.note_int = note_int
            self.note_str = int_to_str(note_int)
        else:
            raise ValueError('Must provide either note_str or note_int')

    def print_all(self, prefix: str = ''):
        print(prefix + "Note: " + self.note_str + ", " + str(self.note_int))


class Mode:

    mode_int: int = None
    mode_str: str = None

    def __init__(self, mode: int = None, mode_str: str = None):
        '''
        mode_int:  0 = Ionian, 1 = Dorian, 2 = Phrygian, 3 = Lydian, 4 = Mixolydian, 5 = Aeolian, 6 = Locrian
        mode_str: 'Ionian', 'Dorian', 'Phrygian', 'Lydian', 'Mixolydian', 'Aeolian', 'Locrian'
        '''
        if mode is not None:
            if mode < 0 or mode > 6:
                raise ValueError('Invalid mode')
            self.mode_int = mode
            self.mode_str = ['Ionian', 'Dorian', 'Phrygian',
                             'Lydian', 'Mixolydian', 'Aeolian', 'Locrian'][mode]
        elif mode_str is not None:
            if mode_str not in ['Ionian', 'Dorian', 'Phrygian', 'Lydian', 'Mixolydian', 'Aeolian', 'Locrian']:
                raise ValueError('Invalid mode_str')
            self.mode_int = ['Ionian', 'Dorian', 'Phrygian', 'Lydian',
                             'Mixolydian', 'Aeolian', 'Locrian'].index(mode_str)
            self.mode_str = mode_str
        else:
            raise ValueError('Must provide either mode or mode_str')

    def print_all(self, prefix: str = ''):
        print(prefix + "Mode: " + self.mode_str + ", " + str(self.mode_int))


class MusicTheory:

    major_scale: List[Note] = None
    natural_minor_scale: List[Note] = None
    chromatic_scale: List[Note] = None
    pentatonic_scale: List[Note] = None
    mode: Mode = None
    modal_scale: List[Note] = None
    relative_minor: Note = None
    relative_major: Note = None
    key: Note = None

    def __init__(self, key_str: str = None, key_int: int = None, mode_int: int = None, mode_str: str = None):
        '''
        key_str: 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'
        #, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        key_int: 0 = C, 1 = C
        mode:  0 = Ionian, 1 = Dorian, 2 = Phrygian, 3 = Lydian, 4 = Mixolydian, 5 = Aeolian, 6 = Locrian
        '''
        if key_str is not None and key_int is not None:
            raise ValueError('Cannot specify both key_str and key_int')
        elif key_str is None and key_int is None:
            raise ValueError('Must specify either key_str or key_int')
        elif key_str is not None:
            self.key = Note(note_str=key_str)
        elif key_int is not None:
            self.key = Note(note_int=key_int)
        else:
            raise ValueError('Invalid key')
        if mode_int is not None and mode_str is not None:
            raise ValueError('Cannot specify both mode_int and mode_str')
        elif mode_int is None and mode_str is None:
            raise ValueError('Must specify either mode_int or mode_str')
        elif mode_str is not None:
            self.mode = Mode(mode_str=mode_str)
        elif mode_int is not None:
            self.mode = Mode(mode=mode_int)
        else:
            raise ValueError('Invalid mode')

        self.major_scale = self.__major_scale()
        self.natural_minor_scale = self.__natural_minor_scale()
        self.chromatic_scale = self.__chromatic_scale()
        self.pentatonic_scale = self.__pentatonic_scale()
        self.modal_scale = self.__modal_scale()
        self.relative_minor = self.__relative_minor()
        self.relative_major = self.__relative_major()

    def __major_scale(self) -> List[Note]:
        '''
        key: 0 = C, 1 = C#, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        '''
        steps = [2, 2, 1, 2, 2, 2, 1]
        notes: List[Note] = [self.key]
        for i in range(len(steps)):
            note_int = (notes[i].note_int + steps[i]) % 12
            notes.append(Note(note_int=note_int))
        return notes

    def __natural_minor_scale(self) -> List[Note]:
        '''
        key: 0 = C, 1 = C#, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        '''
        steps = [2, 1, 2, 2, 1, 2, 2]
        notes: List[Note] = [self.key]
        for i in range(len(steps)):
            note_int = (notes[i].note_int + steps[i]) % 12
            notes.append(Note(note_int=note_int))
        return notes

    def __chromatic_scale(self) -> list:
        '''
        key: 0 = C, 1 = C#, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        '''
        notes: List[Note] = [self.key]
        for i in range(1, 12):
            note_int = (self.key.note_int + i) % 12
            notes.append(Note(note_int=note_int))
        return notes

    def __pentatonic_scale(self) -> list:
        '''
        key: 0 = C, 1 = C#, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        '''
        notes_int = [(self.key.note_int + i*7) % 12 for i in range(5)]
        notes: List[Note] = [Note(note_int=note_int) for note_int in notes_int]
        return notes

    def __modal_scale(self) -> list:
        '''
        key: 0 = C, 1 = C#, 2 = D, 3 = D#, 4 = E, 5 = F, 6 = F#, 7 = G, 8 = G#, 9 = A, 10 = A#, 11 = B
        mode: 0 = Ionian, 1 = Dorian, 2 = Phrygian, 3 = Lydian, 4 = Mixolydian, 5 = Aeolian, 6 = Locrian
        '''
        steps = [2, 2, 1, 2, 2, 1, 2]
        notes: List[Note] = [self.key]
        for i in range(len(steps)):
            note_int = (notes[i].note_int + steps[i]) % 12
            notes.append(Note(note_int=note_int))
        notes = notes[self.mode.mode_int:] + notes[:self.mode.mode_int]
        return notes

    def __relative_minor(self) -> Note:
        """
        Returns the relative minor key for a given major key.

        Parameters:
        key (str): The major key to find the relative minor for.

        Returns:
        str: The relative minor key.
        """

        # Create a dictionary mapping major keys to their relative minor keys
        # use sharps only
        relative_minor_dict = {'C': 'A', 'G': 'E', 'D': 'B', 'A': 'F#', 'E': 'C#', 'B': 'G#', 'F#': 'D#',
                               'F#': 'D#', 'C#': 'A#', 'G#': 'F', 'D#': 'C', 'A#': 'G', 'F': 'D'}

        # Look up the relative minor key and return it
        return Note(note_str=relative_minor_dict[self.key.note_str])

    def __relative_major(self) -> str:
        """
        Returns the relative major key for a given minor key.

        Parameters:
        key (str): The minor key to find the relative major for.

        Returns:
        str: The relative major key.
        """

        # Create a dictionary mapping minor keys to their relative major keys
        relative_major_dict = {'A': 'C', 'E': 'G', 'B': 'D', 'F#': 'A', 'C#': 'E', 'G#': 'B', 'D#': 'F#',
                               'F': 'G#', 'C': 'D#', 'G': 'A#', 'D': 'F', 'A#': 'C#', 'D#': 'F#'}

        # Look up the relative major key and return it
        return Note(note_str=relative_major_dict[self.key.note_str])

    def print_all(self):
        print('key:', self.key.note_str)
        attrs = vars(self)

        for attr, value in attrs.items():
            # if type Note, call Note.print()
            if isinstance(value, Note):
                print('-', attr, ':')
                value.print_all('--')
            # if type Mode, call Note.print()
            elif isinstance(value, Mode):
                print('-', attr, ':')
                value.print_all('--')
            # if type List[Note], call Note.print() for each Note in List
            elif isinstance(value, list):
                print('-', attr, ':')
                if isinstance(value[0], Note):
                    for note in value:
                        note: Note
                        note.print_all('--')
            # if type List[Mode], call Mode.print() for each Mode in List
                elif isinstance(value[0], Mode):
                    for mode in value:
                        mode: Mode
                        mode.print_all('--')
            else:
                print('-', attr, ':', value)


# example

if __name__ == '__main__':

    # create a music_theory object
    for key in ['C', 'G', 'D', 'A', 'E', 'B', 'F#', 'C#', 'G#', 'D#', 'A#', 'F']:
        for mode in ['Ionian', 'Dorian', 'Phrygian', 'Lydian', 'Mixolydian', 'Aeolian', 'Locrian']:
            print('key:', key, 'mode:', mode)
            mt = MusicTheory(key_str=key, mode_str=mode)

            # print all attributes
            mt.print_all()



'''
original code from 'https://github.com/jackmcarthur/musical-key-finder'
modified by: 'bin2ai' to be used as an installable package with pip
'''

import numpy as np
import matplotlib.pyplot as plt
import librosa
import librosa.display

# class that uses the librosa library to analyze the key that an mp3 is in
# arguments:
#     waveform: an mp3 file loaded by librosa, ideally separated out from any percussive sources
#     sr: sampling rate of the mp3, which can be obtained when the file is read with librosa
#     tstart and tend: the range in seconds of the file to be analyzed; default to the beginning and end of file if not specified


class KeyFinder(object):

    filename: str = None  # path to mp3 file
    title: str = None  # title of song (defaults to filename if None)
    waveform: np.ndarray = None  # waveform of song
    duration: float = None  # song duration in seconds
    sr: int = None  # sampling rate
    t_start: int = None  # time in seconds
    t_end: int = None  # time in seconds
    s_start: int = None  # sample number
    s_end: int = None  # sample number
    y_segment: np.ndarray = None  # waveform of song segment
    chromograph: np.ndarray = None  # chroma graph of song segment
    chroma_vals: list = None  # amount of each pitch class present in this time interval
    keyfreqs: dict = None  # pitch names to the associated intensity in the song
    min_key_corrs: list = None  # correlations between pitch class vs minor keys
    maj_key_corrs: list = None  # correlations between pitch class vs major keys
    key_dict: dict = None    # dictionary of the musical keys (major/minor)
    key_primary: str = None  # key determined by the algorithm
    # strength of correlation for the key determined by the algorithm
    best_corr_primary: float = None
    key_alt: str = None  # alternative key determined by the algorithm
    best_corr_alt: float = None  # strength of correlation for the alternative key
    chroma_max: float = None  # maximum value of the chroma graph
    pitch_max: str = None  # pitch class with the highest intensity in the song segment
    pitch_max_corr: float = None  # strength of pitch class with the highest intensity
    pitches: list = None  # list of pitch classes
    # major key profile from Krumhansl-Schmuckler algorithm
    major_profile_ks: list = None
    # minor key profile from Krumhansl-Schmuckler algorithm
    minor_profile_ks: list = None

    def __init__(self,
                 file: str,  # path to mp3 file
                 title: str = None,  # title of song, if None, defaults to filename from path
                 t_start: int = None,  # time in seconds, if None, defaults to beginning of file
                 t_end: int = None  # time in seconds, if None, defaults to end of file
                 ) -> None:

        self.filename = file

        if title is None:
            self.title = file
        else:
            self.title = title

        self.waveform, self.sr = librosa.load(path=file, sr=None)
        self.duration = librosa.get_duration(y=self.waveform, sr=self.sr)

        self.waveform
        self.sr = self.sr

        self.t_start = t_start
        self.t_end = t_end

        if self.t_start is not None:
            self.s_start = librosa.time_to_samples(self.t_start, sr=self.sr)
        if self.t_end is not None:
            self.s_end = librosa.time_to_samples(self.t_end, sr=self.sr)
        self.y_segment = self.waveform[self.s_start:self.s_end]
        self.chromograph = librosa.feature.chroma_cqt(
            y=self.y_segment, sr=self.sr, bins_per_octave=24)

        # chroma_vals is the amount of each pitch class present in this time interval
        self.chroma_vals = []
        for i in range(12):
            self.chroma_vals.append(np.sum(self.chromograph[i]))
        self.pitches = ['C', 'C#', 'D', 'D#', 'E',
                        'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
        # dictionary relating pitch names to the associated intensity in the song
        self.keyfreqs = {self.pitches[i]
            : self.chroma_vals[i] for i in range(12)}

        # list of musical keys (major/minor)
        keys = [
            self.pitches[i] + ' major' for i in range(12)] + [self.pitches[i] + ' minor' for i in range(12)]

        # use of the Krumhansl-Schmuckler key-finding algorithm, which compares the chroma
        # data above to typical profiles of major and minor keys:
        self.major_profile_ks = [6.35, 2.23, 3.48, 2.33, 4.38,
                                 4.09, 2.52, 5.19, 2.39, 3.66, 2.29, 2.88]
        self.minor_profile_ks = [6.33, 2.68, 3.52, 5.38, 2.60,
                                 3.53, 2.54, 4.75, 3.98, 2.69, 3.34, 3.17]

        # finds correlations between the amount of each pitch class in the time interval and the above profiles,
        # starting on each of the 12 pitches. then creates dict of the musical keys (major/minor) to the correlation
        self.min_key_corrs = []
        self.maj_key_corrs = []
        for i in range(12):
            key_test = [self.keyfreqs.get(self.pitches[(i + m) % 12])
                        for m in range(12)]
            # correlation coefficients (strengths of correlation for each key)
            self.maj_key_corrs.append(
                round(np.corrcoef(self.major_profile_ks, key_test)[1, 0], 3))
            self.min_key_corrs.append(
                round(np.corrcoef(self.minor_profile_ks, key_test)[1, 0], 3))

        # names of all major and minor keys
        self.key_dict = {**{keys[i]: self.maj_key_corrs[i] for i in range(12)},
                         **{keys[i+12]: self.min_key_corrs[i] for i in range(12)}}

        # this attribute represents the key determined by the algorithm
        self.key_primary = max(self.key_dict, key=self.key_dict.get)
        self.best_corr_primary = max(self.key_dict.values())

        # this attribute represents the second-best key determined by the algorithm,
        # if the correlation is close to that of the actual key determined
        self.key_alt = None
        self.best_corr_alt = None

        best_corr_alt = 0
        for key, corr in self.key_dict.items():
            # if the correlation is close to the best correlation, but not the same,
            # if corr > self.best_corr_primary*0.9 and corr != self.best_corr_primary:
            #     self.key_alt = key
            #     self.best_corr_alt = corr
            # get the second best key
            if corr < self.best_corr_primary and corr >= best_corr_alt:
                self.key_alt = key
                best_corr_alt = corr
                self.best_corr_alt = best_corr_alt

        # done initializing

    # get primary key and correlation
    def get_primary_key_corr(self) -> tuple:
        return self.key_primary, self.best_corr_primary

    # get secondary key and correlation
    def get_secondary_key_corr(self) -> tuple:
        return self.key_alt, self.best_corr_alt

    # prints the relative prominence of each pitch class
    def print_chroma(self) -> None:
        self.chroma_max = max(self.chroma_vals)
        for key, chrom in self.keyfreqs.items():
            print(key, '\t', f'{chrom/self.chroma_max:5.3f}')

    # prints the correlation coefficients associated with each major/minor key
    def print_corr_table(self) -> None:
        for key, corr in self.key_dict.items():
            print(key, '\t', f'{corr:6.3f}')

    # printout of the key determined by the algorithm; if another key is close, that key is mentioned
    def print_key(self) -> None:
        print("likely key: ", max(self.key_dict, key=self.key_dict.get),
              ", correlation: ", self.best_corr_primary, sep='')
        if self.key_alt is not None:
            print("also possible: ", self.key_alt,
                  ", correlation: ", self.best_corr_alt, sep='')

    # prints a chromagram of the file, showing the intensity of each pitch class over time
    def show_chromagram(self, title: str = 'Chromagram') -> None:
        C = librosa.feature.chroma_cqt(
            y=self.waveform, sr=self.sr, bins_per_octave=24)
        plt.figure(figsize=(12, 4))
        librosa.display.specshow(
            C, sr=self.sr, x_axis='time', y_axis='chroma', vmin=0, vmax=1)
        if title is None:
            plt.title('Chromagram')
        else:
            plt.title(title)
        plt.colorbar()
        plt.tight_layout()
        plt.show()


# example usage:
if __name__ == '__main__':

    file = 'angry-birds-theme-song.mp3'
    song = KeyFinder(file)
    song.print_chroma()
    song.print_corr_table()
    song.print_key()
    song.show_chromagram()

    # print primary key and correlation
    print(song.get_primary_key_corr())
    # print secondary key and correlation
    print(song.get_secondary_key_corr())

    # for attr in dir(song):
    #     if not attr.startswith('_'):
    #         # if not an ndarray, or a list
    #         if not isinstance(getattr(song, attr), np.ndarray):
    #             print(attr, ':', getattr(song, attr))



from setuptools import setup, find_packages
import pathlib

# The directory containing this file
here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name='pymusickit',
    version='0.1.1',
    packages=find_packages(),
    install_requires=[
        'librosa==0.9.2',
        'matplotlib==3.7.0',
        'numpy==1.23.5'
    ],
    entry_points={
        'console_scripts': [
            'key_finder=pymusickit.key_finder:main',
            'music_theory=pymusickit.music_theory:main',
        ],
    },
    description='A Python package for music analysis.  Keyfinder Forked from "https://github.com/jackmcarthur/musical-key-finder"',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='bin2ai',
    url='https://github.com/bin2ai/pymusickit',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6'
)
